"""

merge_detections.py

Merge high-confidence detections from one or more results files into another
file.  Typically used to combine results from MDv5b and/or MDv4 into a "primary"
results file from MDv5a.

Detection categories must be the same in both files; if you want to first remap
one file's category mapping to be the same as another's, see remap_detection_categories.

If you want to literally merge two .json files, see combine_batch_outputs.py.

"""

#%% Constants and imports

import argparse
import sys
import json
import os

from tqdm import tqdm

from megadetector.utils.ct_utils import get_iou


#%% Structs

class MergeDetectionsOptions:
    """
    Class defining options for merge_detections().
    """

    def __init__(self):

        #: Maximum detection size to include in the merged output
        self.max_detection_size = 1.01

        #: Minimum detection size to include in the merged output
        self.min_detection_size = 0

        #: Exclude detections whose confidence in the source file(s) is less
        #: than this.  Should have the same length as the number of source files.
        self.source_confidence_thresholds = [0.05]

        #: Don't bother merging into target images if there is a similar detection
        #: above this threshold (or if there is *any* detection above this threshold,
        #: and merge_empty_only is True)
        self.target_confidence_threshold = 0.2

        #: If you want to merge only certain categories, specify one
        #: (but not both) of these.  These are category IDs, not names.
        self.categories_to_include = None

        #: If you want to merge only certain categories, specify one
        #: (but not both) of these.  These are category IDs, not names.
        self.categories_to_exclude = None

        #: Only merge detections into images that have *no* detections in the
        #: target results file.
        self.merge_empty_only = False

        #: IoU threshold above which two detections are considered the same
        self.iou_threshold = 0.65

        #: Error if this is False and the output file exists
        self.overwrite = False


#%% Main function

def merge_detections(source_files,target_file,output_file,options=None):
    """
    Merge high-confidence detections from one or more results files into another
    file.   Typically used to combine results from MDv5b and/or MDv4 into a "primary"
    results file from MDv5a.

    [source_files] (a list of files or a single filename) specifies the set of
    results files that will be merged into [target_file].  The difference between a
    "source file" and the "target file" is that if no merging is necessary, either because
    two boxes are nearly identical or because merge_only_empty is True and the target
    file already has above-threshold detection for an image+category, the output file gets
    the results of the "target" file.  I.e., the "target" file wins all ties.

    The results are written to [output_file].

    """

    if isinstance(source_files,str):
        source_files = [source_files]

    if options is None:
        options = MergeDetectionsOptions()

    if (not options.overwrite) and (os.path.isfile(output_file)):
        print('File {} exists, bypassing merge'.format(output_file))
        return

    assert not ((options.categories_to_exclude is not None) and \
                (options.categories_to_include is not None)), \
                'categories_to_include and categories_to_exclude are mutually exclusive'

    if options.categories_to_exclude is not None:
        options.categories_to_exclude = [int(c) for c in options.categories_to_exclude]

    if options.categories_to_include is not None:
        options.categories_to_include = [int(c) for c in options.categories_to_include]

    assert len(source_files) == len(options.source_confidence_thresholds), \
        '{} source files provided, but {} source confidence thresholds provided'.format(
            len(source_files),len(options.source_confidence_thresholds))

    for fn in source_files:
        assert os.path.isfile(fn), 'Could not find source file {}'.format(fn)

    assert os.path.isfile(target_file)

    os.makedirs(os.path.dirname(output_file),exist_ok=True)

    with open(target_file,'r') as f:
        output_data = json.load(f)

    print('Loaded results for {} images'.format(len(output_data['images'])))

    fn_to_image = {}

    # im = output_data['images'][0]
    for im in output_data['images']:
        fn_to_image[im['file']] = im

    if 'detections_transferred_from' not in output_data['info']:
        output_data['info']['detections_transferred_from'] = []

    if 'detector' not in output_data['info']:
        output_data['info']['detector'] = 'MDv4 (assumed)'

    detection_categories_raw = output_data['detection_categories'].keys()

    # Determine whether we should be processing all categories, or just a subset
    # of categories.
    detection_categories = []

    if options.categories_to_exclude is not None:
        for c in detection_categories_raw:
            if int(c) not in options.categories_to_exclude:
                detection_categories.append(c)
            else:
                print('Excluding category {}'.format(c))
    elif options.categories_to_include is not None:
        for c in detection_categories_raw:
            if int(c) in options.categories_to_include:
                print('Including category {}'.format(c))
                detection_categories.append(c)
    else:
        detection_categories = detection_categories_raw

    # i_source_file = 0; source_file = source_files[i_source_file]
    for i_source_file,source_file in enumerate(source_files):

        print('Processing detections from file {}'.format(source_file))

        with open(source_file,'r') as f:
            source_data = json.load(f)

        if 'detector' in source_data['info']:
            source_detector_name = source_data['info']['detector']
        else:
            source_detector_name = os.path.basename(source_file)

        output_data['info']['detections_transferred_from'].append(os.path.basename(source_file))
        output_data['info']['detector'] = output_data['info']['detector'] + ' + ' + source_detector_name

        assert source_data['detection_categories'] == output_data['detection_categories'], \
            'Cannot merge files with different detection category maps'

        source_confidence_threshold = options.source_confidence_thresholds[i_source_file]

        # source_im = source_data['images'][0]
        for source_im in tqdm(source_data['images']):

            image_filename = source_im['file']

            assert image_filename in fn_to_image, 'Image {} not in target image set'.format(image_filename)
            target_im = fn_to_image[image_filename]

            if 'detections' not in source_im or source_im['detections'] is None:
                continue

            if 'detections' not in target_im or target_im['detections'] is None:
                continue

            source_detections_this_image = source_im['detections']
            target_detections_this_image = target_im['detections']

            detections_to_transfer = []

            # detection_category = list(detection_categories)[0]
            for detection_category in detection_categories:

                target_detections_this_category = \
                    [det for det in target_detections_this_image if det['category'] == \
                     detection_category]

                max_target_confidence_this_category = 0.0

                if len(target_detections_this_category) > 0:
                    max_target_confidence_this_category = max([det['conf'] for \
                      det in target_detections_this_category])

                # If we have a valid detection in the target file, and we're only merging
                # into images that have no detections at all, we don't need to review the individual
                # detections in the source file.
                if options.merge_empty_only and \
                    (max_target_confidence_this_category >= options.target_confidence_threshold):
                    continue

                source_detections_this_category_raw = [det for det in \
                  source_detections_this_image if det['category'] == detection_category]

                # Boxes are x/y/w/h
                # source_sizes = [det['bbox'][2]*det['bbox'][3] for det in source_detections_this_category_raw]

                # Only look at source boxes within the size range
                source_detections_this_category_filtered = [
                    det for det in source_detections_this_category_raw if \
                        (det['bbox'][2]*det['bbox'][3] <= options.max_detection_size) and \
                        (det['bbox'][2]*det['bbox'][3] >= options.min_detection_size) \
                        ]

                # det = source_detections_this_category_filtered[0]
                for det in source_detections_this_category_filtered:

                    if det['conf'] >= source_confidence_threshold:

                        # Check only whole images
                        if options.merge_empty_only:

                            # We verified this above, asserting here for clarity
                            assert max_target_confidence_this_category < options.target_confidence_threshold
                            det['transferred_from'] = source_detector_name
                            detections_to_transfer.append(det)

                        # Check individual detections
                        else:

                            # Does this source detection match any existing above-threshold
                            # target category detections?
                            matches_existing_box = False

                            # target_detection = target_detections_this_category[0]
                            for target_detection in target_detections_this_category:

                                if (target_detection['conf'] >= options.target_confidence_threshold) \
                                   and \
                                   (get_iou(det['bbox'],target_detection['bbox']) >= options.iou_threshold):
                                    matches_existing_box = True
                                    break

                            if (not matches_existing_box):
                                det['transferred_from'] = source_detector_name
                                detections_to_transfer.append(det)

                    # ...if this source detection is above the confidence threshold

                # ...for each source detection within category

            # ...for each detection category

            if len(detections_to_transfer) > 0:

                # print('Adding {} detections to image {}'.format(len(detections_to_transfer),image_filename))
                detections = fn_to_image[image_filename]['detections']
                detections.extend(detections_to_transfer)

                # Update the max_detection_conf field (if present)
                if 'max_detection_conf' in fn_to_image[image_filename]:
                    fn_to_image[image_filename]['max_detection_conf'] = \
                        max([d['conf'] for d in detections])

            # ...if we have any detections to transfer

        # ...for each image

    # ...for each source file

    with open(output_file,'w') as f:
        json.dump(output_data,f,indent=1)

    print('Saved merged results to {}'.format(output_file))


#%% Command-line driver

def main(): # noqa

    default_options = MergeDetectionsOptions()

    parser = argparse.ArgumentParser(
        description='Merge detections from one or more MegaDetector results files into an existing reuslts file')
    parser.add_argument(
        'source_files',
        nargs="+",
        help='Path to source .json file(s) to merge from')
    parser.add_argument(
        'target_file',
        help='Path to a .json file to merge detections into')
    parser.add_argument(
        'output_file',
        help='Path to output .json results file')
    parser.add_argument(
        '--max_detection_size',
        type=float,
        default=default_options.max_detection_size,
        help='Ignore detections with an area larger than this (as a fraction of ' + \
             'image size) (default {})'.format(
             default_options.max_detection_size))
    parser.add_argument(
        '--min_detection_size',
        default=default_options.min_detection_size,
        type=float,
        help='Ignore detections with an area smaller than this (as a fraction of ' + \
              'image size) (default {})'.format(
              default_options.min_detection_size))
    parser.add_argument(
        '--source_confidence_thresholds',
        nargs="+",
        type=float,
        default=default_options.source_confidence_thresholds,
        help='List of thresholds for each source file (default {}). '.format(
            default_options.source_confidence_thresholds) + \
            'Merge only if the source file\'s detection confidence is higher than its ' + \
            'corresponding threshold.  Should be the same length as the number of source files.')
    parser.add_argument(
        '--target_confidence_threshold',
        type=float,
        default=default_options.target_confidence_threshold,
        help='Don\'t merge if target file\'s detection confidence is already higher ' + \
             'than this (default {}). '.format(
             default_options.target_confidence_threshold))
    parser.add_argument(
        '--categories_to_include',
        type=int,
        nargs="+",
        default=None,
        help='List of numeric detection category IDs to include')
    parser.add_argument(
        '--categories_to_exclude',
        type=int,
        nargs="+",
        default=None,
        help='List of numeric detection categories to include')
    parser.add_argument(
        '--merge_empty_only',
        action='store_true',
        help='Ignore individual detections and only merge images for which the target ' + \
             'file contains no detections')
    parser.add_argument(
        '--iou_threshold',
        type=float,
        default=default_options.iou_threshold,
        help='Sets the minimum IoU for a source detection to be considered the same as ' + \
             'a target detection (default {})'.format(default_options.iou_threshold))

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    options = MergeDetectionsOptions()
    options.max_detection_size = args.max_detection_size
    options.min_detection_size = args.min_detection_size
    options.source_confidence_thresholds = args.source_confidence_thresholds
    options.target_confidence_threshold = args.target_confidence_threshold
    options.categories_to_include = args.categories_to_include
    options.categories_to_exclude = args.categories_to_exclude
    options.merge_empty_only = args.merge_empty_only
    options.iou_threshold = args.iou_threshold

    merge_detections(args.source_files, args.target_file, args.output_file, options)

if __name__ == '__main__':
    main()




#
# If a request has been sent to AML for batch scoring but the monitoring thread of the API was
# interrupted (uncaught exception or having to re-start the API container), we could manually
# aggregate results from each shard using this script, assuming all jobs submitted to AML have finished.
#
# Need to have set environment variables STORAGE_ACCOUNT_NAME and STORAGE_ACCOUNT_KEY to those of the
# storage account backing the API. Also need to adjust the INTERNAL_CONTAINER, AML_CONTAINER and
# AML_CONFIG fields in api_core/orchestrator_api/api_config.py to match the instance of the API that this
# request was submitted to.
#
# May need to change the import statement in api_core/orchestrator_api/orchestrator.py
# "from sas_blob_utils import SasBlob" to
# "from .sas_blob_utils import SasBlob" to not confuse with the module in AI4Eutils;
# and change "import api_config" to
# "from api.batch_processing.api_core.orchestrator_api import api_config"

# Execute this script from the root of the repository. You may need to add the repository to PYTHONPATH.

import argparse
import json

from api.batch_processing.api_core.orchestrator_api.orchestrator import AMLMonitor


def main(): # noqa
    parser = argparse.ArgumentParser()
    parser.add_argument('shortened_request_id', type=str,
                        help='the request ID to restart monitoring')
    parser.add_argument('model_version', type=str, help='version of megadetector used; this is used to fill in the meta info section of the output file')
    parser.add_argument('request_name', type=str, help='easy to remember name for that job, optional', default='')
    args = parser.parse_args()


    # list_jobs_submitted cannot be serialized ("can't pickle _thread.RLock objects "), but
    # do not need it for aggregating results
    aml_monitor = AMLMonitor(request_id=args.request_id,
                             list_jobs_submitted=None,
                             request_name=args.request_name,
                             request_submission_timestamp='',
                             model_version=args.model_version)
    output_file_urls = aml_monitor.aggregate_results()
    output_file_urls_str = json.dumps(output_file_urls)
    print(output_file_urls_str)

if __name__ == '__main__':
    main()


# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""
A class to manage updating the status of an API request / Azure Batch Job using
the Cosmos DB table "batch_api_jobs".
"""

import logging
import os
import unittest
import uuid
from typing import Union, Optional

from azure.cosmos.cosmos_client import CosmosClient
from azure.cosmos.exceptions import CosmosResourceNotFoundError

from server_api_config import API_INSTANCE_NAME, COSMOS_ENDPOINT, COSMOS_WRITE_KEY
from server_utils import get_utc_time


log = logging.getLogger(os.environ['FLASK_APP'])


class JobStatusTable:
    """
    A wrapper around the Cosmos DB client. Each item in the table "batch_api_jobs" represents
    a request/Batch Job, and should have the following fields:
        - id: this is the job_id
        - api_instance
        - status
        - last_updated
        - call_params: the dict representing the body of the POST request from the user
    The 'status' field is a dict with the following fields:
        - request_status
        - message
        - num_tasks  (present after Batch Job created)
        - num_images  (present after Batch Job created)
    """
    # a job moves from created to running/problem after the Batch Job has been submitted
    allowed_statuses = ['created', 'running', 'failed', 'problem', 'completed', 'canceled']

    def __init__(self, api_instance=None):
        self.api_instance = api_instance if api_instance is not None else API_INSTANCE_NAME
        cosmos_client = CosmosClient(COSMOS_ENDPOINT, credential=COSMOS_WRITE_KEY)
        db_client = cosmos_client.get_database_client('camera-trap')
        self.db_jobs_client = db_client.get_container_client('batch_api_jobs')

    def create_job_status(self, job_id: str, status: Union[dict, str], call_params: dict) -> dict:
        assert 'request_status' in status and 'message' in status
        assert status['request_status'] in JobStatusTable.allowed_statuses

        # job_id should be unique across all instances, and is also the partition key
        cur_time = get_utc_time()
        item = {
            'id': job_id,
            'api_instance': self.api_instance,
            'status': status,
            'job_submission_time': cur_time,
            'last_updated': cur_time,
            'call_params': call_params
        }
        created_item = self.db_jobs_client.create_item(item)
        return created_item

    def update_job_status(self, job_id: str, status: Union[dict, str]) -> dict:
        assert 'request_status' in status and 'message' in status
        assert status['request_status'] in JobStatusTable.allowed_statuses

        item_old = self.read_job_status(job_id)
        if item_old is None:
            raise ValueError

        # need to retain other fields in 'status' to be able to restart monitoring thread
        if 'status' in item_old and isinstance(item_old['status'], dict):
            # retain existing fields; update as needed
            for k, v in item_old['status'].items():
                if k not in status:
                    status[k] = v
        item = {
            'id': job_id,
            'api_instance': self.api_instance,
            'status': status,
            'job_submission_time': item_old['job_submission_time'],
            'last_updated': get_utc_time(),
            'call_params': item_old['call_params']
        }
        replaced_item = self.db_jobs_client.replace_item(job_id, item)
        return replaced_item

    def read_job_status(self, job_id) -> Optional[dict]:
        """
        Read the status of the job from the Cosmos DB table of job status.
        Note that it does not check the actual status of the job on Batch, and just returns what
        the monitoring thread wrote to the database.
        job_id is also the partition key
        """
        try:
            read_item = self.db_jobs_client.read_item(job_id, partition_key=job_id)
            assert read_item['api_instance'] == self.api_instance, 'Job does not belong to this API instance'
        except CosmosResourceNotFoundError:
            return None  # job_id not a key
        except Exception as e:
            logging.error(f'server_job_status_table, read_job_status, exception: {e}')
            raise
        else:
            item = {k: v for k, v in read_item.items() if not k.startswith('_')}
            return item


class TestJobStatusTable(unittest.TestCase):
    api_instance = 'api_test'

    def test_insert(self):
        table = JobStatusTable(TestJobStatusTable.api_instance)
        status = {
            'request_status': 'running',
            'message': 'this is a test'
        }
        job_id = uuid.uuid4().hex
        item = table.create_job_status(job_id, status, {'container_sas': 'random_string'})
        self.assertTrue(job_id == item['id'], 'Expect job_id to be the id of the item')
        self.assertTrue(item['status']['request_status'] == 'running', 'Expect fields to be inserted correctly')

    def test_update_and_read(self):
        table = JobStatusTable(TestJobStatusTable.api_instance)
        status = {
            'request_status': 'running',
            'message': 'this is a test'
        }
        job_id = uuid.uuid4().hex
        res = table.create_job_status(job_id, status, {'container_sas': 'random_string'})

        status = {
            'request_status': 'completed',
            'message': 'this is a test again'
        }
        res = table.update_job_status(job_id, status)
        item_read = table.read_job_status(job_id)
        self.assertTrue(item_read['status']['request_status'] == 'completed', 'Expect field to have updated')

    def test_read_invalid_id(self):
        table = JobStatusTable(TestJobStatusTable.api_instance)
        job_id = uuid.uuid4().hex  # should not be in the database
        item_read = table.read_job_status(job_id)
        self.assertIsNone(item_read)

if __name__ == '__main__':
    unittest.main()



"""

csv_to_json.py

Converts CSV to JSON format for label specification.

There are 3 possible values for the 'type' column in the CSV:

- "row": this selects a specific rowfrom the master taxonomy CSV
    content syntax: <dataset_name>|<dataset_label>

- "datasettaxon": this selects all animals in a taxon from a particular dataset
    content syntax: <dataset_name>|<taxon_level>|<taxon_name>

- <taxon_level>: this selects all animals in a taxon across all datasets
    content syntax: <taxon_name>

Example CSV input:

"
  # comment lines starting with '#' are allowed
  output_label,type,content
  cervid,row,idfg|deer
  cervid,row,idfg|elk
  cervid,row,idfg|prong
  cervid,row,idfg_swwlf_2019|elk
  cervid,row,idfg_swwlf_2019|muledeer
  cervid,row,idfg_swwlf_2019|whitetaileddeer
  cervid,max_count,50000
  cervid,family,cervidae
  cervid,datasettaxon,idfg|family|cervidae
  cervid,datasettaxon,idfg_swwlf_2019|family|cervidae
  bird,row,idfg_swwlf_2019|bird
  bird,class,aves
  bird,max_count,50000
  bird,prioritize,"[['idfg_swwlf_2019'], ['idfg']]"
  !bird,row,idfg_swwlf_2019|turkey
  !bird,genus,meleagris
"

Example JSON output:

"
  {
      "cervid": {
          "dataset_labels": {
              "idfg": ["deer", "elk", "prong"],
              "idfg_swwlf_2019": ["elk", "muledeer", "whitetaileddeer"]
          },
          "taxa": [
              {
                  "level": "family",
                  "name": "cervidae"
              },
              {
                  "level": "family",
                  "name": "cervidae"
                  "datasets": ["idfg"]
              },
              {
                  "level": "family",
                  "name": "cervidae"
                  "datasets": ["idfg_swwlf_2019"]
              }
          ],
          "max_count": 50000
      },
      "bird": {
          "dataset_labels": {
              "idfg_swwlf_2019": ["bird"]
          },
          "taxa": [
              {
                  "level": "class",
                  "name": "aves"
              }
          ],
          "exclude": {
              "dataset_labels": {
                  "idfg_swwlf_2019": ["turkey"]
              },
              "taxa": [
                  {
                      "level": "genus",
                      "name": "meleagris"
                  }
              ]
          },
          "max_count": "50000",
          "prioritize": [
              ["idfg_swwlf_2019"],
              ["idfg"]
          ],
      }
  }
"

"""

#%% Imports

from __future__ import annotations

import argparse
from collections import defaultdict
import json
from typing import Any

import pandas as pd
from megadetector.utils import ct_utils


#%% Main function

def main(): # noqa
    args = _parse_args()
    js = csv_to_jsondict(args.input_csv_file)
    for label in js:
        js[label] = order_spec_dict(js[label])
    ct_utils.write_json(args.output_json_path, js, indent=args.json_indent)


#%% Support functions

def parse_csv_row(obj: dict[str, Any], rowtype: str, content: str) -> None:
    """
    Parses a row in the CSV.
    """

    if rowtype == 'row':
        if 'dataset_labels' not in obj:
            obj['dataset_labels'] = defaultdict(list)

        assert '|' in content
        ds, ds_label = content.split('|')
        obj['dataset_labels'][ds].append(ds_label)

    elif rowtype == 'datasettaxon':
        if 'taxa' not in obj:
            obj['taxa'] = []

        assert '|' in content
        ds, taxon_level, taxon_name = content.split('|')
        obj['taxa'].append({
            'level': taxon_level,
            'name': taxon_name,
            'datasets': [ds]
        })

    elif rowtype == 'max_count':
        obj['max_count'] = int(content)

    elif rowtype == 'prioritize':
        obj['prioritize'] = eval(content)  # pylint: disable=eval-used

    else:
        if 'taxa' not in obj:
            obj['taxa'] = []

        taxon_level = rowtype
        taxon_name = content
        obj['taxa'].append({
            'level': taxon_level,
            'name': taxon_name
        })


def csv_to_jsondict(csv_path: str) -> dict[str, dict[str, Any]]:
    """
    Converts CSV to json-style dictionary.
    """

    df = pd.read_csv(csv_path, comment='#', skip_blank_lines=True)
    assert (df.columns == ['output_label', 'type', 'content']).all()

    js: defaultdict[str, dict[str, Any]] = defaultdict(dict)

    for i in df.index:
        label, rowtype, content = df.loc[i]
        if label.startswith('!'):
            label = label[1:]
            if 'exclude' not in js[label]:
                js[label]['exclude'] = {}
            obj = js[label]['exclude']
        else:
            obj = js[label]
        parse_csv_row(obj, rowtype, content)

    return js


def order_spec_dict(spec_dict: dict[str, Any]) -> dict[str, Any]:
    """
    Returns spec_dict with keys in a specific order.
    """

    if 'exclude' in spec_dict:
        spec_dict['exclude'] = order_spec_dict(spec_dict['exclude'])
    ordered_spec_dict: dict[str, Any] = {}
    for key in ['dataset_labels', 'taxa', 'exclude', 'max_count', 'prioritize']:
        if key in spec_dict:
            ordered_spec_dict[key] = spec_dict[key]
    return ordered_spec_dict


#%% Command-line driver

def _parse_args() -> argparse.Namespace:

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Converts CSV to JSON format for label specification.')
    parser.add_argument(
        'input_csv_file',
        help='path to CSV file containing label specification')
    parser.add_argument(
        'output_json_path',
        help='path to save converted JSON file')
    parser.add_argument(
        '--json-indent', type=int,
        help='number of spaces to use for JSON indent (default no indent)')
    return parser.parse_args()


if __name__ == '__main__':
    main()



"""

sas_blob_utils.py

This module contains helper functions for dealing with Shared Access Signatures
(SAS) tokens for Azure Blob Storage.

The default Azure Storage SAS URI format is:

https://<account>.blob.core.windows.net/<container>/<blob>?<sas_token>

This module assumes azure-storage-blob version 12.5.

Documentation for Azure Blob Storage:
docs.microsoft.com/en-us/azure/developer/python/sdk/storage/storage-blob-readme

Documentation for SAS:
docs.microsoft.com/en-us/azure/storage/common/storage-sas-overview

"""

#%% Imports

from datetime import datetime, timedelta
import io
import re
from typing import (Any, AnyStr, IO, Iterable, Optional, Union)
from urllib import parse
import uuid

from tqdm import tqdm

from azure.storage.blob import ( # type: ignore
    BlobClient,
    BlobProperties,
    ContainerClient,
    ContainerSasPermissions,
    generate_container_sas)


#%% URI management

def build_azure_storage_uri(
        account: str,
        container: Optional[str] = None,
        blob: Optional[str] = None,
        sas_token: Optional[str] = None,
        account_url_template: str = 'https://{account}.blob.core.windows.net'
        ) -> str:
    """
    Build an Azure SAS URL from path name components.

    Args:
        account: str, name of Azure Storage account
        container: optional str, name of Azure Blob Storage container
        blob: optional str, name of blob, not URL-escaped
            if blob is given, must also specify container
        sas_token: optional str, Shared Access Signature (SAS).  Leading ?
            is removed if present.
        account_url_template: str, Python 3 string formatting template,
            contains '{account}' placeholder, defaults to default Azure
            Storage URL format. Set this value if using Azurite Azure Storage
            emulator.

    Returns:
        str, Azure storage URI
    """

    uri = account_url_template.format(account=account)
    if container is not None:
        uri = f'{uri}/{container}'
    if blob is not None:
        assert container is not None
        blob = parse.quote(blob)
        uri = f'{uri}/{blob}'
    if sas_token is not None:
        if sas_token[0] == '?':
            sas_token = sas_token[1:]
        uri = f'{uri}?{sas_token}'
    return uri


def _get_resource_reference(prefix: str) -> str:
    return '{}{}'.format(prefix, str(uuid.uuid4()).replace('-', ''))


def get_client_from_uri(container_uri: str) -> ContainerClient:
    """
    Gets a ContainerClient for the given container URI.
    """

    return ContainerClient.from_container_url(container_uri)


def get_account_from_uri(sas_uri: str) -> str:
    """
    Assumes that sas_uri points to Azure Blob Storage account hosted at
    a default Azure URI. Does not work for locally-emulated Azure Storage
    or Azure Storage hosted at custom endpoints.
    """

    url_parts = parse.urlsplit(sas_uri)
    loc = url_parts.netloc  # "<account>.blob.core.windows.net"
    return loc.split('.')[0]


def is_container_uri(sas_uri: str) -> bool:
    """
    Returns True if the signed resource field in the URI "sr" is a container "c"
    or a directory "d"
    """

    data = get_all_query_parts(sas_uri)
    if 'sr' not in data:
        return False

    if 'c' in data['sr'] or 'd' in data['sr']:
        return True
    else:
        return False

def is_blob_uri(sas_uri: str) -> bool:
    """
    Returns True if the signed resource field in the URI "sr" is a blob "b".
    """

    data = get_all_query_parts(sas_uri)
    if 'sr' not in data:
        return False

    if 'b' in data['sr']:
        return True
    else:
        return False


def get_container_from_uri(sas_uri: str, unquote: bool = True) -> str:
    """
    Gets the container name from a Azure Blob Storage URI.

    Assumes that sas_uri points to Azure Blob Storage account hosted at
    a default Azure URI. Does not work for locally-emulated Azure Storage
    or Azure Storage hosted at custom endpoints.

    Args:
        sas_uri: str, Azure blob storage URI, may include SAS token
        unquote: bool, whether to replace any %xx escapes by their
            single-character equivalent, default True

    Returns:
        str, container name
    """

    url_parts = parse.urlsplit(sas_uri)
    raw_path = url_parts.path.lstrip('/')  # remove leading "/" from path
    container = raw_path.split('/')[0]
    if container == '':
        raise ValueError('Given sas_uri does not include a container.')
    if unquote:
        container = parse.unquote(container)
    return container


def get_blob_from_uri(sas_uri: str, unquote: bool = True) -> str:
    """
    Return the path to the blob from the root container if this sas_uri
    is for an individual blob; otherwise returns None.

    Args:
        sas_uri: str, Azure blob storage URI, may include SAS token
        unquote: bool, whether to replace any %xx escapes by their
            single-character equivalent, default True

    Returns: str, blob name (path to the blob from the root container)

    Raises: ValueError, if sas_uri does not include a blob name
    """

    # Get the entire path with all slashes after the container
    url_parts = parse.urlsplit(sas_uri)
    raw_path = url_parts.path.lstrip('/')  # remove leading "/" from path
    parts = raw_path.split('/', maxsplit=1)
    if len(parts) < 2 or parts[1] == '':
        raise ValueError('Given sas_uri does not include a blob name')

    blob = parts[1]  # first item is an empty string
    if unquote:
        blob = parse.unquote(blob)
    return blob


def get_sas_token_from_uri(sas_uri: str) -> Optional[str]:
    """
    Get the query part of the SAS token that contains permissions, access
    times and signature.

    Args:
        sas_uri: str, Azure blob storage SAS token

    Returns: str, query part of the SAS token (without leading '?'),
        or None if URI has no token.
    """

    url_parts = parse.urlsplit(sas_uri)
    sas_token = url_parts.query or None  # None if query is empty string
    return sas_token


def get_resource_type_from_uri(sas_uri: str) -> Optional[str]:
    """
    Get the resource type pointed to by this SAS token.

    Args:
        sas_uri: str, Azure blob storage URI with SAS token

    Returns: A string (either 'blob' or 'container') or None.
    """

    url_parts = parse.urlsplit(sas_uri)
    data = parse.parse_qs(url_parts.query)
    if 'sr' in data:
        types = data['sr']
        if 'b' in types:
            return 'blob'
        elif 'c' in types:
            return 'container'
    return None


def get_endpoint_suffix(sas_uri):
    """
    Gets the endpoint at which the blob storage account is served.

    Args:
        sas_uri: str, Azure blob storage URI with SAS token

    Returns:
        str: usually 'core.windows.net' or 'core.chinacloudapi.cn', to
         use for the `endpoint` argument in various blob storage SDK functions.
    """

    url_parts = parse.urlsplit(sas_uri)
    suffix = url_parts.netloc.split('.blob.')[1].split('/')[0]
    return suffix


def get_permissions_from_uri(sas_uri: str) -> set[str]:
    """
    Get the permissions given by this SAS token.

    Args:
        sas_uri: str, Azure blob storage URI with SAS token

    Returns:
        set: a set containing some of 'read', 'write', 'delete' and 'list'.
        Empty set returned if no permission specified in sas_uri.
    """

    data = get_all_query_parts(sas_uri)
    permissions_set = set()
    if 'sp' in data:
        permissions = data['sp'][0]
        if 'r' in permissions:
            permissions_set.add('read')
        if 'w' in permissions:
            permissions_set.add('write')
        if 'd' in permissions:
            permissions_set.add('delete')
        if 'l' in permissions:
            permissions_set.add('list')
    return permissions_set


def get_all_query_parts(sas_uri: str) -> dict[str, Any]:
    """
    Gets the SAS token parameters.
    """

    url_parts = parse.urlsplit(sas_uri)
    return parse.parse_qs(url_parts.query)


#%% Blob

def check_blob_exists(sas_uri: str, blob_name: Optional[str] = None) -> bool:
    """
    Checks whether a given URI points to an actual blob.

    Assumes that sas_uri points to Azure Blob Storage account hosted at
    a default Azure URI. Does not work for locally-emulated Azure Storage
    or Azure Storage hosted at custom endpoints. In these cases, create a
    BlobClient using the default constructor, instead of from_blob_url(),
    and use the BlobClient.exists() method directly.

    Args:
        sas_uri: str, URI to a container or a blob
            if blob_name is given, sas_uri is treated as a container URI
            otherwise, sas_uri is treated as a blob URI
        blob_name: optional str, name of blob, not URL-escaped
            must be given if sas_uri is a URI to a container

    Returns:
        bool: whether the sas_uri given points to an existing blob
    """

    if blob_name is not None:
        sas_uri = build_blob_uri(
            container_uri=sas_uri, blob_name=blob_name)

    with BlobClient.from_blob_url(sas_uri) as blob_client:
        return blob_client.exists()


def upload_blob(container_uri: str, blob_name: str,
                data: Union[Iterable[AnyStr], IO[AnyStr]],
                overwrite: bool = False) -> str:
    """
    Creates a new blob of the given name from an IO stream.

    Args:
        container_uri: str, URI to a container, may include SAS token
        blob_name: str, name of blob to upload
        data: str, bytes, or IO stream
            if str, assumes utf-8 encoding
        overwrite: bool, whether to overwrite existing blob (if any)

    Returns:
        str: URL to blob, includes SAS token if container_uri has SAS token
    """

    account_url, container, sas_token = split_container_uri(container_uri)
    with BlobClient(account_url=account_url, container_name=container,
                    blob_name=blob_name, credential=sas_token) as blob_client:
        blob_client.upload_blob(data, overwrite=overwrite)
        return blob_client.url


def download_blob_to_stream(sas_uri: str) -> tuple[io.BytesIO, BlobProperties]:
    """
    Downloads a blob to an IO stream.

    Args:
        sas_uri: str, URI to a blob

    Returns:
        output_stream: io.BytesIO, remember to close it when finished using
        blob_properties: BlobProperties
    """

    with BlobClient.from_blob_url(sas_uri) as blob_client:
        output_stream = io.BytesIO()
        blob_client.download_blob().readinto(output_stream)
        output_stream.seek(0)
        blob_properties = blob_client.get_blob_properties()
    return output_stream, blob_properties


def build_blob_uri(container_uri: str, blob_name: str) -> str:
    """
    Args:
        container_uri: str, URI to blob storage container
            <account_url>/<container>?<sas_token>
        blob_name: str, name of blob, not URL-escaped

    Returns:
        str: blob URI <account_url>/<container>/<blob_name>?<sas_token>,
        <blob_name> is URL-escaped
    """

    account_url, container, sas_token = split_container_uri(container_uri)

    blob_name = parse.quote(blob_name)
    blob_uri = f'{account_url}/{container}/{blob_name}'
    if sas_token is not None:
        blob_uri += f'?{sas_token}'
    return blob_uri


#%% Container

def list_blobs_in_container(
        container_uri: str,
        blob_prefix: Optional[str] = None,
        blob_suffix: Optional[Union[str, tuple[str]]] = None,
        rsearch: Optional[str] = None,
        limit: Optional[int] = None,
        verbose: Optional[bool] = True
) -> list[str]:
    """
    Get a sorted list of blob names in this container.

    Args:
        container_uri: str, URI to a container, may include SAS token
        blob_prefix: optional str, returned results will only contain blob names
            to with this prefix
        blob_suffix: optional str or tuple of str, returned results will only
            contain blob names with this/these suffix(es). The blob names will
            be lowercased first before comparing with the suffix(es).
        rsearch: optional str, returned results will only contain blob names
            that match this regex. Can also be a list of regexes, in which case
            blobs matching *any* of the regex's will be returned.
        limit: int, maximum # of blob names to list
            if None, then returns all blob names

    Returns:
        list: sorted list of blob names, of length limit or shorter.
    """

    if verbose:
        print('Listing blobs')

    if (get_sas_token_from_uri(container_uri) is not None
            and get_resource_type_from_uri(container_uri) != 'container'):
        raise ValueError('The SAS token provided is not for a container.')

    if blob_prefix is not None and not isinstance(blob_prefix, str):
        raise ValueError('blob_prefix must be a str.')

    if (blob_suffix is not None
            and not isinstance(blob_suffix, str)
            and not isinstance(blob_suffix, tuple)):
        raise ValueError('blob_suffix must be a str or a tuple of strings')

    list_blobs = []
    with get_client_from_uri(container_uri) as container_client:
        generator = container_client.list_blobs(
            name_starts_with=blob_prefix)

        if blob_suffix is None and rsearch is None:
            list_blobs = [blob.name for blob in tqdm(generator,disable=(not verbose))]
            i = len(list_blobs)
        else:
            i = 0
            for blob in tqdm(generator,disable=(not verbose)):
                i += 1
                suffix_ok = (blob_suffix is None
                             or blob.name.lower().endswith(blob_suffix))
                regex_ok = False
                if rsearch is None:
                    regex_ok = True
                else:
                    if not isinstance(rsearch, list):
                        rsearch = [rsearch]
                    # Check whether this blob name matches *any* of our regex's
                    for expr in rsearch:
                        if re.search(expr, blob.name) is not None:
                            regex_ok = True
                            break
                if suffix_ok and regex_ok:
                    list_blobs.append(blob.name)
                    if limit is not None and len(list_blobs) == limit:
                        break

    if verbose:
        print(f'Enumerated {len(list_blobs)} matching blobs out of {i} total')

    return sorted(list_blobs)  # sort for determinism


def generate_writable_container_sas(account_name: str,
                                    account_key: str,
                                    container_name: str,
                                    access_duration_hrs: float,
                                    account_url: Optional[str] = None
                                    ) -> str:
    """
    Creates a container and returns a SAS URI with read/write/list
    permissions.

    Args:
        account_name: str, name of blob storage account
        account_key: str, account SAS token or account shared access key
        container_name: str, name of container to create, must not match an
            existing container in the given storage account
        access_duration_hrs: float
        account_url: str, optional, defaults to default Azure Storage URL

    Returns:
        str: URL to newly created container
    """

    if account_url is None:
        account_url = build_azure_storage_uri(account=account_name)
    with ContainerClient(account_url=account_url,
                         container_name=container_name,
                         credential=account_key) as container_client:
        container_client.create_container()

    permissions = ContainerSasPermissions(read=True, write=True, list=True)
    container_sas_token = generate_container_sas(
        account_name=account_name,
        container_name=container_name,
        account_key=account_key,
        permission=permissions,
        expiry=datetime.utcnow() + timedelta(hours=access_duration_hrs))

    return f'{account_url}/{container_name}?{container_sas_token}'


def split_container_uri(container_uri: str) -> tuple[str, str, Optional[str]]:
    """
    Args:
        container_uri: str, URI to blob storage container
            <account_url>/<container>?<sas_token>

    Returns:
        tuple: account_url, container_name, sas_token
    """

    account_container = container_uri.split('?', maxsplit=1)[0]
    account_url, container_name = account_container.rsplit('/', maxsplit=1)
    sas_token = get_sas_token_from_uri(container_uri)
    return account_url, container_name, sas_token



"""

detector_calibration.py

Tools for comparing/calibrating confidence values from detectors, particularly different
versions of MegaDetector.

"""

#%% Constants and imports

import random
import copy

from tqdm import tqdm
from enum import IntEnum
from collections import defaultdict

import numpy as np
import matplotlib
import matplotlib.pyplot as plt

from megadetector.postprocessing.validate_batch_results import \
    validate_batch_results, ValidateBatchResultsOptions
from megadetector.utils.ct_utils import get_iou, max_none, is_iterable


#%% Classes

class CalibrationOptions:
    """
    Options controlling comparison/calibration behavior.
    """

    def __init__(self):

        #: IoU threshold used for determining whether two detections are the same
        #:
        #: When multiple detections match, we will only use the highest-matching IoU.
        self.iou_threshold = 0.6

        #: Minimum confidence threshold to consider for calibration (should be lower than
        #: the lowest value you would use in realistic situations)
        self.confidence_threshold = 0.025

        #: Should we populate the data_a and data_b fields in the return value?
        self.return_data = False

        #: Model name to use in printouts and plots for result set A
        self.model_name_a = 'model_a'

        #: Model name to use in printouts and plots for result set B
        self.model_name_b = 'model_b'

        #: Maximum number of samples to use for plotting or calibration per category,
        #: or None to use all paired values.  If separate_plots_by_category is False,
        #: this is the overall number of points sampled.
        self.max_samples_per_category = None

        #: Should we make separate plots for each category?  Mutually exclusive with
        #: separate_plots_by_correctness.
        self.separate_plots_by_category = True

        #: Should we make separate plots for TPs/FPs?  Mutually exclusive with
        #: separate_plots_by_category.
        self.separate_plots_by_correctness = False

        #: List of category IDs to use for plotting comparisons, or None to plot
        #: all categories.
        self.categories_to_plot = None

        #: Optionally map category ID to name in plot labels
        self.category_id_to_name = None

        #: Enable additional debug output
        self.verbose = True

# ...class CalibrationOptions

class CalibrationMatchColumns(IntEnum):
    """
    Enumeration defining columns in the calibration_matches list we'll assemble below.
    """

    COLUMN_CONF_A = 0
    COLUMN_CONF_B = 1
    COLUMN_IOU = 2
    COLUMN_I_IMAGE = 3
    COLUMN_CATEGORY_ID = 4
    COLUMN_MATCHES_GT = 5

class CalibrationResults:
    """
    Results of a model-to-model comparison.
    """

    def __init__(self):

        #: List of tuples: [conf_a, conf_b, iou, i_image, category_id, matches_gt]
        #:
        #: If ground truth is supplied, [matches_gt] is a bool indicating whether either
        #: of the detected boxes matches a ground truth box of the same category.  If
        #: ground truth is not supplied, [matches_gt] is None.
        self.calibration_matches = []

        #: Populated with the data loaded from json_filename_a if options.return_data is True
        self.data_a = None

        #: Populated with the data loaded from json_filename_b if options.return_data is True
        self.data_b = None

# ...class CalibrationResults


#%% Calibration functions

def compare_model_confidence_values(json_filename_a,json_filename_b,json_filename_gt=None,options=None):
    """
    Compare confidence values across two .json results files.  Compares only detections that
    can be matched by IoU, i.e., does not do anything with detections that only appear in one file.

    Args:
        json_filename_a (str or dict): filename containing results from the first model to be compared;
            should refer to the same images as [json_filename_b].  Can also be a loaded results dict.
        json_filename_b (str or dict): filename containing results from the second model to be compared;
            should refer to the same images as [json_filename_a].  Can also be a loaded results dict.
        json_filename_gt (str or dict, optional): filename containing ground truth; should refer to the
            same images as [json_filename_a] and [json_filename_b].  Can also be a loaded results dict.
            Should be in COCO format.
        options (CalibrationOptions, optional): all the parameters used to control this process, see
            CalibrationOptions for details

    Returns:
        CalibrationResults: description of the comparison results
    """

    ## Option handling

    if options is None:
        options = CalibrationOptions()

    validation_options = ValidateBatchResultsOptions()
    validation_options.return_data = True

    if isinstance(json_filename_a,str):
        results_a = validate_batch_results(json_filename_a,options=validation_options)
        assert len(results_a['validation_results']['errors']) == 0
    else:
        assert isinstance(json_filename_a,dict)
        results_a = json_filename_a

    if isinstance(json_filename_b,str):
        results_b = validate_batch_results(json_filename_b,options=validation_options)
        assert len(results_b['validation_results']['errors']) == 0
    else:
        assert isinstance(json_filename_b,dict)
        results_b = json_filename_b

    # Load ground truth, if supplied
    gt_data = None

    if json_filename_gt is not None:
        if isinstance(json_filename_gt,str):
            gt_data = validate_batch_results(json_filename_gt,
                                             options=validation_options)
        else:
            assert isinstance(json_filename_gt,dict)
            gt_data = json_filename_gt

    ## Make sure these results sets are comparable

    image_filenames_a = [im['file'] for im in results_a['images']]
    image_filenames_b = [im['file'] for im in results_b['images']]

    assert set(image_filenames_a) == set(image_filenames_b), \
        'Cannot calibrate non-matching image sets'

    categories_a = results_a['detection_categories']
    categories_b = results_b['detection_categories']
    assert set(categories_a.keys()) == set(categories_b.keys())
    for k in categories_a.keys():
        assert categories_a[k] == categories_b[k], 'Category mismatch'


    ## Load ground truth if necessary

    gt_category_name_to_id = None
    gt_image_id_to_annotations = None
    image_filename_to_gt_im = None

    if gt_data is not None:

        gt_category_name_to_id = {}
        for c in gt_data['categories']:
            gt_category_name_to_id[c['name']] = c['id']

        image_filename_to_gt_im = {}
        for im in gt_data['images']:
            assert 'width' in im and 'height' in im, \
                'I can only compare against GT that has "width" and "height" fields'
            image_filename_to_gt_im[im['file_name']] = im

        assert set(image_filename_to_gt_im.keys()) == set(image_filenames_a), \
            'Ground truth filename list does not match image filename list'

        gt_image_id_to_annotations = defaultdict(list)
        for ann in gt_data['annotations']:
            gt_image_id_to_annotations[ann['image_id']].append(ann)


    ## Compare detections

    image_filename_b_to_im = {}
    for im in results_b['images']:
        image_filename_b_to_im[im['file']] = im

    n_detections_a = 0
    n_detections_a_queried = 0
    n_detections_a_matched = 0

    calibration_matches = []

    # For each image
    # im_a = results_a['images'][0]
    for i_image,im_a in tqdm(enumerate(results_a['images']),total=len(results_a['images'])):

        fn = im_a['file']
        im_b = image_filename_b_to_im[fn]

        if 'detections' not in im_a or im_a['detections'] is None:
            continue
        if 'detections' not in im_b or im_b['detections'] is None:
            continue

        im_gt = None
        if gt_data is not None:
            im_gt = image_filename_to_gt_im[fn]

        # For each detection in result set A...
        #
        # det_a = im_a['detections'][0]
        for det_a in im_a['detections']:

            n_detections_a += 1

            conf_a = det_a['conf']
            category_id = det_a['category']

            # Is this above threshold?
            if conf_a < options.confidence_threshold:
                continue

            n_detections_a_queried += 1

            bbox_a = det_a['bbox']

            best_iou = None
            best_iou_conf = None
            best_bbox_b = None

            # For each detection in result set B...
            #
            # det_b = im_b['detections'][0]
            for det_b in im_b['detections']:

                # Is this the same category?
                if det_b['category'] != category_id:
                    continue

                conf_b = det_b['conf']

                # Is this above threshold?
                if conf_b < options.confidence_threshold:
                    continue

                bbox_b = det_b['bbox']

                iou = get_iou(bbox_a,bbox_b)

                # Is this an adequate IoU to consider?
                if iou < options.iou_threshold:
                    continue

                # Is this the best match so far?
                if best_iou is None or iou > best_iou:
                    best_iou = iou
                    best_iou_conf = conf_b
                    best_bbox_b = bbox_b

            # ...for each detection in im_b

            # If we found a match between A and B
            if best_iou is not None:

                n_detections_a_matched += 1

                # Does this pair of matched detections also match a ground truth box?
                matches_gt = None

                if im_gt is not None:

                    def max_iou_between_detection_and_gt(detection_box,category_name,im_gt,gt_annotations):

                        max_iou = None

                        # Which category ID are we looking for?
                        gt_category_id_for_detected_category_name = \
                            gt_category_name_to_id[category_name]

                        # For each GT annotation
                        #
                        # ann = gt_annotations[0]
                        for ann in gt_annotations:

                            # Only match against boxes in the same category
                            if ann['category_id'] != gt_category_id_for_detected_category_name:
                                continue
                            if 'bbox' not in ann:
                                continue

                            # Normalize this box
                            #
                            # COCO format: [x,y,width,height]
                            # normalized format: [x_min, y_min, width_of_box, height_of_box]
                            normalized_gt_box = [ann['bbox'][0]/im_gt['width'],ann['bbox'][1]/im_gt['height'],
                                                 ann['bbox'][2]/im_gt['width'],ann['bbox'][3]/im_gt['height']]

                            iou = get_iou(detection_box, normalized_gt_box)
                            if max_iou is None or iou > max_iou:
                                max_iou = iou

                        # ...for each gt box

                        return max_iou

                    # ...def min_iou_between_detections_and_gt(...)

                    gt_annotations = gt_image_id_to_annotations[im_gt['id']]

                    # If they matched, the A and B boxes have the same category by definition
                    category_name = categories_a[det_a['category']]

                    max_iou_with_bbox_a = \
                        max_iou_between_detection_and_gt(bbox_a,category_name,im_gt,gt_annotations)
                    max_iou_with_bbox_b = \
                        max_iou_between_detection_and_gt(best_bbox_b,category_name,im_gt,gt_annotations)

                    max_iou_with_either_detection_set = max_none(max_iou_with_bbox_a,
                                                                 max_iou_with_bbox_b)

                    matches_gt = False
                    if (max_iou_with_either_detection_set is not None) and \
                       (max_iou_with_either_detection_set >= options.iou_threshold):
                        matches_gt = True

                # ...if we have ground truth

                conf_result = [conf_a,best_iou_conf,best_iou,i_image,category_id,matches_gt]
                calibration_matches.append(conf_result)

            # ...if we had a match between A and B
        # ...for each detection in im_a

    # ...for each image in result set A

    if options.verbose:

        print('\nOf {} detections in result set A, queried {}, matched {}'.format(
            n_detections_a,n_detections_a_queried,n_detections_a_matched))

        if gt_data is not None:
            n_matches = 0
            for m in calibration_matches:
               assert m[CalibrationMatchColumns.COLUMN_MATCHES_GT] is not None
               if m[CalibrationMatchColumns.COLUMN_MATCHES_GT]:
                   n_matches += 1
            print('{} matches also matched ground truth'.format(n_matches))

    assert len(calibration_matches) == n_detections_a_matched

    calibration_results = CalibrationResults()
    calibration_results.calibration_matches = calibration_matches

    if options.return_data:
        calibration_results.data_a = results_a
        calibration_results.data_b = results_b

    return calibration_results

# ...def compare_model_confidence_values(...)


#%% Plotting functions

def plot_matched_confidence_values(calibration_results,output_filename,options=None):
    """
    Given a set of paired confidence values for matching detections (from
    compare_model_confidence_values), plot histograms of those pairs for each
    detection category.

    Args:
        calibration_results (CalibrationResults): output from a call to
            compare_model_confidence_values, containing paired confidence
            values for two sets of detection results.
        output_filename (str): filename to write the plot (.png or .jpg)
        options (CalibrationOptions, optional): plotting options, see
            CalibrationOptions for details.
    """

    fig_w = 12
    fig_h = 8
    n_hist_bins = 80

    if options is None:
        options = CalibrationOptions()

    assert not (options.separate_plots_by_category and \
                options.separate_plots_by_correctness), \
        'separate_plots_by_category and separate_plots_by_correctness are mutually exclusive'

    category_id_to_name = None
    category_to_samples = None

    calibration_matches = calibration_results.calibration_matches

    # If we're just lumping everything into one plot
    if (not options.separate_plots_by_category) and (not options.separate_plots_by_correctness):

        category_id_to_name = {'0':'all_categories'}
        category_to_samples = {'0': []}

        # Make everything category "0" (arbitrary)
        calibration_matches = copy.deepcopy(calibration_matches)
        for m in calibration_matches:
            m[CalibrationMatchColumns.COLUMN_CATEGORY_ID] = '0'
        if (options.max_samples_per_category is not None) and \
           (len(calibration_matches) > options.max_samples_per_category):
            calibration_matches = \
                random.sample(calibration_matches,options.max_samples_per_category)
        category_to_samples['0'] = calibration_matches

    # If we're separating into lines for FPs and TPs (but not separating by category)
    elif options.separate_plots_by_correctness:

        assert not options.separate_plots_by_category

        category_id_tp = '0'
        category_id_fp = '1'

        category_id_to_name = {category_id_tp:'TP', category_id_fp:'FP'}
        category_to_samples = {category_id_tp: [], category_id_fp: []}

        for m in calibration_matches:
            assert m[CalibrationMatchColumns.COLUMN_MATCHES_GT] is not None, \
                "Can't plot by correctness when GT status is not available for every match"
            if m[CalibrationMatchColumns.COLUMN_MATCHES_GT]:
                category_to_samples[category_id_tp].append(m)
            else:
                category_to_samples[category_id_fp].append(m)

    # If we're separating by category
    else:

        assert options.separate_plots_by_category

        category_to_samples = defaultdict(list)

        category_to_matches = defaultdict(list)
        for m in calibration_matches:
            category_id = m[CalibrationMatchColumns.COLUMN_CATEGORY_ID]
            category_to_matches[category_id].append(m)

        category_id_to_name = None
        if options.category_id_to_name is not None:
            category_id_to_name = options.category_id_to_name

        for i_category,category_id in enumerate(category_to_matches.keys()):

            matches_this_category = category_to_matches[category_id]

            if (options.max_samples_per_category is None) or \
                (len(matches_this_category) <= options.max_samples_per_category):
                category_to_samples[category_id] = matches_this_category
            else:
                assert len(matches_this_category) > options.max_samples_per_category
                category_to_samples[category_id] = random.sample(matches_this_category,options.max_samples_per_category)

        del category_to_matches

    del calibration_matches

    if options.verbose:
        n_samples_for_histogram = 0
        for c in category_to_samples:
            n_samples_for_histogram += len(category_to_samples[c])
        print('Creating a histogram based on {} samples'.format(n_samples_for_histogram))

    categories_to_plot = list(category_to_samples.keys())

    if options.categories_to_plot is not None:
        categories_to_plot = [category_id for category_id in categories_to_plot if\
                              category_id in options.categories_to_plot]

    n_subplots = len(categories_to_plot)

    plt.ioff()

    fig = matplotlib.figure.Figure(figsize=(fig_w, fig_h), tight_layout=True)
    # fig,axes = plt.subplots(nrows=n_subplots,ncols=1)

    axes = fig.subplots(n_subplots, 1)

    if not is_iterable(axes):
        assert n_subplots == 1
        axes = [axes]

    # i_category = 0; category_id = categories_to_plot[i_category]
    for i_category,category_id in enumerate(categories_to_plot):

        ax = axes[i_category]

        category_string = str(category_id)
        if (category_id_to_name is not None) and (category_id in category_id_to_name):
            category_string = category_id_to_name[category_id]

        samples_this_category = category_to_samples[category_id]
        x = [m[0] for m in samples_this_category]
        y = [m[1] for m in samples_this_category]

        weights_a = np.ones_like(x)/float(len(x))
        weights_b = np.ones_like(y)/float(len(y))

        # Plot the first lie a little thicker so the second line will always show up
        ax.hist(x,histtype='step',bins=n_hist_bins,density=False,color='red',weights=weights_a,linewidth=3.0)
        ax.hist(y,histtype='step',bins=n_hist_bins,density=False,color='blue',weights=weights_b,linewidth=1.5)

        ax.legend([options.model_name_a,options.model_name_b])
        ax.set_ylabel(category_string)
        # plt.tight_layout()

        # I experimented with heat maps, but they weren't very informative.
        # Leaving this code here in case I revisit.  Note to self: scatter plots
        # were a disaster.
        if False:
            heatmap, xedges, yedges = np.histogram2d(x, y, bins=30)
            extent = [xedges[0], xedges[-1], yedges[0], yedges[-1]]
            plt.imshow(heatmap.T, extent=extent, origin='lower', norm='log')

    # ...for each category for which we need to generate a histogram

    plt.close(fig)
    fig.savefig(output_filename,dpi=100)

# ...def plot_matched_confidence_values(...)


#%% Interactive driver(s)

if False:

    #%%

    options = ValidateBatchResultsOptions()
    # json_filename = r'g:\temp\format.json'
    # json_filename = r'g:\temp\test-videos\video_results.json'
    json_filename = r'g:\temp\test-videos\image_results.json'
    options.check_image_existence = True
    options.relative_path_base = r'g:\temp\test-videos'
    validate_batch_results(json_filename,options)




# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""
This script can be run in a separate process to monitor all instances of the batch API.
It sends a digest of submissions within the past day to a Teams channel webhook.

It requires the environment variables TEAMS_WEBHOOK, COSMOS_ENDPOINT and COSMOS_READ_KEY to be set.
"""

import time
import os
import json
from datetime import datetime, timezone, timedelta
from collections import defaultdict

import requests
from azure.cosmos.cosmos_client import CosmosClient


# Cosmos DB `batch-api-jobs` table for job status
COSMOS_ENDPOINT = os.environ['COSMOS_ENDPOINT']
COSMOS_READ_KEY = os.environ['COSMOS_READ_KEY']

TEAMS_WEBHOOK = os.environ['TEAMS_WEBHOOK']


def send_message():
    cosmos_client = CosmosClient(COSMOS_ENDPOINT, credential=COSMOS_READ_KEY)
    db_client = cosmos_client.get_database_client('camera-trap')
    db_jobs_client = db_client.get_container_client('batch_api_jobs')

    yesterday = datetime.now(timezone.utc).date() - timedelta(days=1)

    query = f'''
    SELECT *
    FROM job
    WHERE job.job_submission_time >= "{yesterday.isoformat()}T00:00:00Z"
    '''

    result_iterable = db_jobs_client.query_items(query=query,
                                                 enable_cross_partition_query=True)

    # aggregate the number of images, country and organization names info from each job
    # submitted during yesterday (UTC time)
    instance_num_images = defaultdict(lambda: defaultdict(int))
    instance_countries = defaultdict(set)
    instance_orgs = defaultdict(set)

    total_images_received = 0

    for job in result_iterable:
        api_instance = job['api_instance']
        status = job['status']
        call_params = job['call_params']

        if status['request_status'] == 'completed':
            instance_num_images[api_instance]['num_images_completed'] += status.get('num_images', 0)
        instance_num_images[api_instance]['num_images_total'] += status.get('num_images', 0)
        total_images_received += status.get('num_images', 0)

        instance_countries[api_instance].add(call_params.get('country', 'unknown'))
        instance_orgs[api_instance].add(call_params.get('organization_name', 'unknown'))

    print(f'send_message, number of images received yesterday: {total_images_received}')

    if total_images_received < 1:
        print('send_message, no images submitted yesterday, not sending a summary')
        print('')
        return

    # create the card
    sections = []

    for instance_name, num_images in instance_num_images.items():
        entry = {
            'activityTitle': f'API instance: {instance_name}',
            'facts': [
                {
                    'name': 'Total images',
                    'value': '{:,}'.format(num_images['num_images_total'])
                },
                {
                    'name': 'Images completed',
                    'value': '{:,}'.format(num_images['num_images_completed'])
                },
                {
                    'name': 'Countries',
                    'value': ', '.join(sorted(list(instance_countries[instance_name])))
                },
                {
                    'name': 'Organizations',
                    'value': ', '.join(sorted(list(instance_orgs[instance_name])))
                }
            ]
        }
        sections.append(entry)

    card = {
        '@type': 'MessageCard',
        '@context': 'http://schema.org/extensions',
        'themeColor': 'ffcdb2',
        'summary': 'Digest of batch API activities over the past 24 hours',
        'title': f'Camera traps batch API activities on {yesterday.strftime("%b %d, %Y")}',
        'sections': sections,
        'potentialAction': [
            {
                '@type': 'OpenUri',
                'name': 'View Batch account in Azure Portal',
                'targets': [
                    {
                        'os': 'default',
                        'uri': 'https://ms.portal.azure.com/#@microsoft.onmicrosoft.com/resource/subscriptions/74d91980-e5b4-4fd9-adb6-263b8f90ec5b/resourcegroups/camera_trap_api_rg/providers/Microsoft.Batch/batchAccounts/cameratrapssc/accountOverview'
                    }
                ]
            }
        ]
    }

    response = requests.post(TEAMS_WEBHOOK, data=json.dumps(card))
    print(f'send_message, card to send:')
    print(json.dumps(card, indent=4))
    print(f'send_message, sent summary to webhook, response code: {response.status_code}')
    print('')


def main(): # noqa
    """
    Wake up at 5 minutes past midnight UTC to send a summary of yesterday's activities if there were any.
    Then goes in a loop to wake up and send a summary every 24 hours.
    """
    current = datetime.utcnow()
    future = current.replace(day=current.day, hour=0, minute=5, second=0, microsecond=0) + timedelta(
        days=1)  # current has been modified

    current = datetime.utcnow()
    duration = future - current

    duration_hours = duration.seconds / (60 * 60)
    print(f'Current time: {current}')
    print(f'Will wake up at {future}, in {duration_hours} hours')
    print('')

    time.sleep(duration.seconds)

    while True:
        print(f'Woke up at {datetime.utcnow()}')
        send_message()
        time.sleep(24 * 60 * 60)

if __name__ == '__main__':
    main()



"""

combine_batch_outputs.py

Merges two or more .json files in MD output format, optionally
writing the results to another .json file.

* Concatenates image lists, erroring if images are not unique.
* Errors if class lists are conflicting; errors on unrecognized fields.
* Checks compatibility in info structs, within reason.

File format:

https://github.com/agentmorris/MegaDetector/tree/main/megadetector/api/batch_processing#batch-processing-api-output-format

Command-line use:

combine_batch_outputs input1.json input2.json ... inputN.json output.json

This does no checking for redundancy; if you are looking to ensemble
the results of multiple model versions, see merge_detections.py.

"""

#%% Constants and imports

import argparse
import sys
import json
from megadetector.utils import ct_utils


#%% Merge functions

def combine_batch_output_files(input_files,
                               output_file=None,
                               require_uniqueness=True,
                               verbose=True):
    """
    Merges the list of MD results files [input_files] into a single
    dictionary, optionally writing the result to [output_file].

    Args:
        input_files (list of str): paths to JSON detection files
        output_file (str, optional): path to write merged JSON
        require_uniqueness (bool): whether to require that the images in
            each list of images be unique

    Returns:
        dict: merged dictionaries loaded from [input_files], identical to what's
        written to [output_file] if [output_file] is not None
    """

    def print_if_verbose(s):
        if verbose:
            print(s)

    input_dicts = []
    for fn in input_files:
        print_if_verbose('Loading results from {}'.format(fn))
        with open(fn, 'r', encoding='utf-8') as f:
            input_dicts.append(json.load(f))

    print_if_verbose('Merging results')
    merged_dict = combine_batch_output_dictionaries(
        input_dicts, require_uniqueness=require_uniqueness)

    print_if_verbose('Writing output to {}'.format(output_file))
    if output_file is not None:
        ct_utils.write_json(output_file, merged_dict)

    return merged_dict


def combine_batch_output_dictionaries(input_dicts, require_uniqueness=True):
    """
    Merges the list of MD results dictionaries [input_dicts] into a single dict.
    See module header comment for details on merge rules.

    Args:
        input_dicts (list of dicts): list of dicts in which each dict represents the
            contents of a MD output file
        require_uniqueness (bool): whether to require that the images in
            each input dict be unique; if this is True and image filenames are
            not unique, an error is raised.

    Returns:
        dict: merged MD results
    """

    # Map image filenames to detections, we'll convert to a list later
    images = {}
    info = {}
    detection_categories = {}
    classification_categories = {}
    n_redundant_images = 0
    n_images = 0

    known_fields = ['info', 'detection_categories', 'classification_categories',
                    'images']

    for input_dict in input_dicts:

        for k in input_dict:
            if k not in known_fields:
                print(f'Warning: unrecognized batch output field: {k}')

        # Check compatibility of detection categories
        for cat_id in input_dict['detection_categories']:
            cat_name = input_dict['detection_categories'][cat_id]
            if cat_id in detection_categories:
                assert detection_categories[cat_id] == cat_name, (
                    'Detection category mismatch')
            else:
                detection_categories[cat_id] = cat_name

        # Check compatibility of classification categories
        if 'classification_categories' in input_dict:
            for cat_id in input_dict['classification_categories']:
                cat_name = input_dict['classification_categories'][cat_id]
                if cat_id in classification_categories:
                    assert classification_categories[cat_id] == cat_name, (
                        'Classification category mismatch')
                else:
                    classification_categories[cat_id] = cat_name

        # Merge image lists, checking uniqueness
        for im in input_dict['images']:
            # Normalize path separators so we don't treat images as different if they
            # were processed on different OS's
            im['file'] = im['file'].replace('\\','/')
            im_file = im['file']
            if require_uniqueness:
                assert im_file not in images, f'Duplicate image: {im_file}'
                images[im_file] = im
                n_images += 1
            else:
                if im_file in images:
                    n_redundant_images += 1
                    previous_im = images[im_file]
                    # Replace a previous failure with a success
                    if ('detections' in im) and ('detections' not in previous_im):
                        images[im_file] = im
                        print(f'Replacing previous failure for image: {im_file}')
                else:
                    images[im_file] = im
                    n_images += 1

        # Merge info dicts, don't check completion time fields
        if len(info) == 0:
            info = input_dict['info']
        else:
            info_compare = input_dict['info']
            assert info_compare['detector'] == info['detector'], (
                'Incompatible detection versions in merging')
            assert info_compare['format_version'] == info['format_version'], (
                'Incompatible batch output versions in merging')
            if 'classifier' in info_compare:
                if 'classifier' in info:
                    assert info['classifier'] == info_compare['classifier']
                else:
                    info['classifier'] = info_compare['classifier']

    # ...for each dictionary

    if n_redundant_images > 0:
        print(f'Warning: found {n_redundant_images} redundant images '
              f'(out of {n_images} total) during merge')

    # Convert merged image dictionaries to a sorted list
    sorted_images = sorted(images.values(), key=lambda im: im['file'])

    merged_dict = {'info': info,
                   'detection_categories': detection_categories,
                   'classification_categories': classification_categories,
                   'images': sorted_images}
    return merged_dict

# ...combine_batch_output_files()


def combine_api_shard_files(input_files, output_file=None):
    """
    Merges the list of .json-formatted API shard files [input_files] into a single
    list of dictionaries, optionally writing the result to [output_file].

    This operates on mostly-deprecated API shard files, not MegaDetector results files.
    If you don't know what an API shard file is, you don't want this function.

    Args:
        input_files (list of str): files to merge
        output_file (str, optional): file to which we should write merged results

    Returns:
        dict: merged results

    :meta private:
    """

    input_lists = []
    print('Loading input files')
    for fn in input_files:
        input_lists.append(json.load(open(fn)))

    detections = []
    # detection_list = input_lists[0]
    for detection_list in input_lists:
        assert isinstance(detection_list, list)
        # d = detection_list[0]
        for d in detection_list:
            assert 'file' in d
            assert 'max_detection_conf' in d
            assert 'detections' in d
            detections.extend([d])

    print('Writing output')
    if output_file is not None:
        ct_utils.write_json(output_file, detections)

    return detections

# ...combine_api_shard_files()


#%% Command-line driver

def main(): # noqa

    parser = argparse.ArgumentParser()
    parser.add_argument(
        'input_paths', nargs='+',
        help='List of input .json files')
    parser.add_argument(
        'output_path',
        help='Output .json file')

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()
    combine_batch_output_files(args.input_paths, args.output_path)

if __name__ == '__main__':
    main()



"""

pytorch_detector.py

Module to run YOLO-based MegaDetector models.

"""

#%% Imports and constants

import os
import sys
import math
import zipfile
import tempfile
import shutil
import traceback
import uuid
import json
import inspect

import cv2
import torch
import numpy as np

from megadetector.detection.run_detector import \
    CONF_DIGITS, COORD_DIGITS, FAILURE_INFER, \
    get_detector_version_from_model_file, \
    known_models
from megadetector.utils.ct_utils import parse_bool_string
from megadetector.utils.ct_utils import to_bool
from megadetector.utils import ct_utils

# We support a few ways of accessing the YOLOv5 dependencies:
#
# * The standard configuration as of 2023.09 expects that the YOLOv5 repo is checked
#   out and on the PYTHONPATH (import utils)
#
# * Supported but non-default (used for PyPI packaging):
#
#   pip install ultralytics-yolov5
#
# * Works, but not supported:
#
#   pip install yolov5
#
# * Unfinished:
#
#   pip install ultralytics

yolo_model_type_imported = None

def _get_model_type_for_model(model_file,
                              prefer_model_type_source='table',
                              default_model_type='yolov5',
                              verbose=False):
    """
    Determine the model type (i.e., the inference library we need to use) for a .pt file.

    Args:
        model_file (str): the model file to read
        prefer_model_type_source (str, optional): how should we handle the (very unlikely)
            case where the metadata in the file indicates one model type, but the global model
            type table says something else.  Should be "table" (trust the table) or "file"
            (trust the file).
        default_model_type (str, optional): return value for the case where we can't find
            appropriate metadata in the file or in the global table.
        verbose (bool, optional): enable additional debug output

    Returns:
        str: the model type indicated for this model
    """

    model_info = read_metadata_from_megadetector_model_file(model_file)

    # Check whether the model file itself specified a model type
    model_type_from_model_file_metadata = None

    if model_info is not None and 'model_type' in model_info:
        model_type_from_model_file_metadata = model_info['model_type']
        if verbose:
            print('Parsed model type {} from model {}'.format(
                model_type_from_model_file_metadata,
                model_file))

    model_type_from_model_version = None

    # Check whether this is a known model version with a specific model type
    model_version_from_file = get_detector_version_from_model_file(model_file)

    if model_version_from_file is not None and model_version_from_file in known_models:
        model_info = known_models[model_version_from_file]
        if 'model_type' in model_info:
            model_type_from_model_version = model_info['model_type']
            if verbose:
                print('Parsed model type {} from global metadata'.format(model_type_from_model_version))
        else:
            model_type_from_model_version = None

    if model_type_from_model_file_metadata is None and \
        model_type_from_model_version is None:
        if verbose:
            print('Could not determine model type for {}, assuming {}'.format(
                model_file,default_model_type))
        model_type = default_model_type

    elif model_type_from_model_file_metadata is not None and \
         model_type_from_model_version is not None:
        if model_type_from_model_version == model_type_from_model_file_metadata:
            model_type = model_type_from_model_file_metadata
        else:
            print('Warning: model type from model version is {}, from file metadata is {}'.format(
                model_type_from_model_version,model_type_from_model_file_metadata))
            if prefer_model_type_source == 'table':
                model_type = model_type_from_model_file_metadata
            else:
                model_type = model_type_from_model_version

    elif model_type_from_model_file_metadata is not None:

        model_type = model_type_from_model_file_metadata

    elif model_type_from_model_version is not None:

        model_type = model_type_from_model_version

    return model_type

# ...def _get_model_type_for_model(...)


def _initialize_yolo_imports_for_model(model_file,
                                       prefer_model_type_source='table',
                                       default_model_type='yolov5',
                                       detector_options=None,
                                       verbose=False):
    """
    Initialize the appropriate YOLO imports for a model file.

    Args:
        model_file (str): The model file for which we're loading support
        prefer_model_type_source (str, optional): how should we handle the (very unlikely)
            case where the metadata in the file indicates one model type, but the global model
            type table says something else.  Should be "table" (trust the table) or "file"
            (trust the file).
        default_model_type (str, optional): return value for the case where we can't find
            appropriate metadata in the file or in the global table.
        detector_options (dict, optional): dictionary of detector options that mean
            different things to different models
        verbose (bool, optional): enable additional debug output

    Returns:
        str: the model type for which we initialized support
    """

    global yolo_model_type_imported

    if detector_options is not None and 'model_type' in detector_options:
        model_type = detector_options['model_type']
        print('Model type {} provided in detector options'.format(model_type))
    else:
        model_type = _get_model_type_for_model(model_file,
                                               prefer_model_type_source=prefer_model_type_source,
                                               default_model_type=default_model_type)

    if yolo_model_type_imported is not None:
        if model_type == yolo_model_type_imported:
            print('Bypassing imports for model type {}'.format(model_type))
            return
        else:
            print('Previously set up imports for model type {}, re-importing as {}'.format(
                yolo_model_type_imported,model_type))

    _initialize_yolo_imports(model_type,verbose=verbose)

    return model_type


def _clean_yolo_imports(verbose=False):
    """
    Remove all YOLO-related imports from sys.modules and sys.path, to allow a clean re-import
    of another YOLO library version.  The reason we jump through all these hoops, rather than
    just, e.g., handling different libraries in different modules, is that we need to make sure
    *pickle* sees the right version of modules during module loading, including modules we don't
    load directly (i.e., every module loaded within a YOLO library), and the only way I know to
    do that is to remove all the "wrong" versions from sys.modules and sys.path.

    Args:
        verbose (bool, optional): enable additional debug output
    """

    modules_to_delete = []
    for module_name in sys.modules.keys():
        module = sys.modules[module_name]
        try:
            module_file = module.__file__.replace('\\','/')
            if 'site-packages' not in module_file:
                continue
            tokens = module_file.split('/')[-4:]
            for token in tokens:
                if 'yolov5' in token or 'yolov9' in token or 'ultralytics' in token:
                    modules_to_delete.append(module_name)
                    break
        except Exception:
            pass

    for module_name in modules_to_delete:
        if module_name in sys.modules.keys():
            module_file = module.__file__.replace('\\','/')
            if verbose:
                print('clean_yolo_imports: deleting module {}: {}'.format(module_name,module_file))
            del sys.modules[module_name]

    paths_to_delete = []

    for p in sys.path:
        if p.endswith('yolov5') or p.endswith('yolov9') or p.endswith('ultralytics'):
            print('clean_yolo_imports: removing {} from path'.format(p))
            paths_to_delete.append(p)

    for p in paths_to_delete:
        sys.path.remove(p)

# ...def _clean_yolo_imports(...)


def _initialize_yolo_imports(model_type='yolov5',
                             allow_fallback_import=True,
                             force_reimport=False,
                             verbose=True):
    """
    Imports required functions from one or more yolo libraries (yolov5, yolov9,
    ultralytics, targeting support for [model_type]).

    Args:
        model_type (str): The model type for which we're loading support
        allow_fallback_import (bool, optional): If we can't import from the package for
            which we're trying to load support, fall back to "import utils".  This is
            typically used when the right support library is on the current PYTHONPATH.
        force_reimport (bool, optional): import the appropriate libraries even if the
            requested model type matches the current initialization state
        verbose (bool, optional): include additional debug output

    Returns:
        str: the model type for which we initialized support
    """

    # When running in pytest, the megadetector 'utils' module is put in the global
    # namespace, which creates conflicts with yolov5; remove it from the global 
    # namespsace.
    if ('PYTEST_CURRENT_TEST' in os.environ):
        print('*** pytest detected ***')
        if ('utils' in sys.modules):
            utils_module = sys.modules['utils']
            if hasattr(utils_module, '__file__') and 'megadetector' in str(utils_module.__file__):
                print(f"Removing conflicting utils module: {utils_module.__file__}")
                sys.modules.pop('utils', None)
                # Also remove any submodules
                to_remove = [name for name in sys.modules if name.startswith('utils.')]
                for name in to_remove:
                    sys.modules.pop(name, None)

    global yolo_model_type_imported

    if model_type is None:
        model_type = 'yolov5'

    # The point of this function is to make the appropriate version
    # of the following functions available at module scope
    global non_max_suppression
    global xyxy2xywh
    global letterbox
    global scale_coords

    if yolo_model_type_imported is not None:
        if (yolo_model_type_imported == model_type) and (not force_reimport):
            print('Bypassing imports for YOLO model type {}'.format(model_type))
            return
        else:
            _clean_yolo_imports()

    try_yolov5_import = (model_type == 'yolov5')
    try_yolov9_import = (model_type == 'yolov9')
    try_ultralytics_import = (model_type == 'ultralytics')

    utils_imported = False

    # First try importing from the yolov5 package; this is how the pip
    # package finds YOLOv5 utilities.
    if try_yolov5_import and not utils_imported:

        try:
            from yolov5.utils.general import non_max_suppression, xyxy2xywh # noqa
            from yolov5.utils.augmentations import letterbox # noqa
            try:
                from yolov5.utils.general import scale_boxes as scale_coords
            except Exception:
                from yolov5.utils.general import scale_coords
            utils_imported = True
            if verbose:
                print('Imported utils from YOLOv5 package')

        except Exception as e: # noqa
            # print('yolov5 module import failed: {}'.format(e))
            # print(traceback.format_exc())
            pass

    # Next try importing from the yolov9 package
    if try_yolov9_import and not utils_imported:

        try:

            from yolov9.utils.general import non_max_suppression, xyxy2xywh # noqa
            from yolov9.utils.augmentations import letterbox # noqa
            from yolov9.utils.general import scale_boxes as scale_coords # noqa
            utils_imported = True
            if verbose:
                print('Imported utils from YOLOv9 package')

        except Exception as e: # noqa

            # print('yolov9 module import failed: {}'.format(e))
            # print(traceback.format_exc())
            pass

    # If we haven't succeeded yet, import from the ultralytics package
    if try_ultralytics_import and not utils_imported:

        try:

            import ultralytics # noqa

        except Exception:

            print('It looks like you are trying to run a model that requires the ultralytics package, '
                  'but the ultralytics package is not installed, but .  For licensing reasons, this '
                  'is not installed by default with the MegaDetector Python package.  Run '
                  '"pip install ultralytics" to install it, and try again.')
            raise

        try:

            from ultralytics.utils.ops import non_max_suppression # noqa
            from ultralytics.utils.ops import xyxy2xywh # noqa

            # In the ultralytics package, scale_boxes and scale_coords both exist;
            # we want scale_boxes.
            #
            # from ultralytics.utils.ops import scale_coords # noqa
            from ultralytics.utils.ops import scale_boxes as scale_coords # noqa
            from ultralytics.data.augment import LetterBox

            # letterbox() became a LetterBox class in the ultralytics package.  Create a
            # backwards-compatible letterbox function wrapper that wraps the class up.
            def letterbox(img,new_shape,auto=False,scaleFill=False, #noqa
                          scaleup=True,center=True,stride=32):

                # Ultralytics changed the "scaleFill" parameter to "scale_fill", we want to support
                # both conventions.
                use_old_scalefill_arg = False
                try:
                    sig = inspect.signature(LetterBox.__init__)
                    if 'scaleFill' in sig.parameters:
                        use_old_scalefill_arg = True
                except Exception:
                    pass

                if use_old_scalefill_arg:
                    if verbose:
                        print('Using old scaleFill calling convention')
                    letterbox_transformer = LetterBox(new_shape,auto=auto,scaleFill=scaleFill,
                                                    scaleup=scaleup,center=center,stride=stride)
                else:
                    letterbox_transformer = LetterBox(new_shape,auto=auto,scale_fill=scaleFill,
                                                  scaleup=scaleup,center=center,stride=stride)
                    
                letterbox_result = letterbox_transformer(image=img)

                if isinstance(new_shape,int):
                    new_shape = [new_shape,new_shape]

                # The letterboxing is done, we just need to reverse-engineer what it did
                shape = img.shape[:2]

                r = min(new_shape[0] / shape[0], new_shape[1] / shape[1])
                if not scaleup:
                    r = min(r, 1.0)
                ratio = r, r

                new_unpad = int(round(shape[1] * r)), int(round(shape[0] * r))
                dw, dh = new_shape[1] - new_unpad[0], new_shape[0] - new_unpad[1]
                if auto:
                    dw, dh = np.mod(dw, stride), np.mod(dh, stride)
                elif scaleFill:
                    dw, dh = 0.0, 0.0
                    new_unpad = (new_shape[1], new_shape[0])
                    ratio = (new_shape[1] / shape[1], new_shape[0] / shape[0])

                dw /= 2
                dh /= 2
                pad = (dw,dh)

                return [letterbox_result,ratio,pad]

            utils_imported = True
            if verbose:
                print('Imported utils from ultralytics package')

        except Exception:

            # print('Ultralytics module import failed')
            pass

    # If we haven't succeeded yet, assume the YOLOv5 repo is on our PYTHONPATH.
    if (not utils_imported) and allow_fallback_import:

        try:

            # import pre- and post-processing functions from the YOLOv5 repo
            from utils.general import non_max_suppression, xyxy2xywh # noqa
            from utils.augmentations import letterbox # noqa

            # scale_coords() is scale_boxes() in some YOLOv5 versions
            try:
                from utils.general import scale_coords # noqa
            except ImportError:
                from utils.general import scale_boxes as scale_coords
            utils_imported = True
            imported_file = sys.modules[scale_coords.__module__].__file__
            if verbose:
                print('Imported utils from {}'.format(imported_file))

        except ModuleNotFoundError as e:

            raise ModuleNotFoundError('Could not import YOLOv5 functions:\n{}'.format(str(e)))

    assert utils_imported, 'YOLO utils import error'

    yolo_model_type_imported = model_type
    if verbose:
        print('Prepared YOLO imports for model type {}'.format(model_type))

    return model_type

# ...def _initialize_yolo_imports(...)


#%% Model metadata functions

def add_metadata_to_megadetector_model_file(model_file_in,
                                            model_file_out,
                                            metadata,
                                            destination_path='megadetector_info.json'):
    """
    Adds a .json file to the specified MegaDetector model file containing metadata used
    by this module.  Always over-writes the output file.

    Args:
        model_file_in (str): The input model filename, typically .pt (.zip is also sensible)
        model_file_out (str): The output model filename, typically .pt (.zip is also sensible).
            May be the same as model_file_in.
        metadata (dict): The metadata dict to add to the output model file
        destination_path (str, optional): The relative path within the main folder of the
            model archive where we should write the metadata.  This is not relative to the root
            of the archive, it's relative to the one and only folder at the root of the archive
            (this is a PyTorch convention).
    """

    tmp_base = os.path.join(tempfile.gettempdir(),'md_metadata')
    os.makedirs(tmp_base,exist_ok=True)
    metadata_tmp_file_relative = 'megadetector_info_' + str(uuid.uuid1()) + '.json'
    metadata_tmp_file_abs = os.path.join(tmp_base,metadata_tmp_file_relative)

    with open(metadata_tmp_file_abs,'w') as f:
        json.dump(metadata,f,indent=1)

    # Copy the input file to the output file
    shutil.copyfile(model_file_in,model_file_out)

    # Write metadata to the output file
    with zipfile.ZipFile(model_file_out, 'a', compression=zipfile.ZIP_DEFLATED) as zipf:

        # Torch doesn't like anything in the root folder of the zipfile, so we put
        # it in the one and only folder.
        names = zipf.namelist()
        root_folders = set()
        for name in names:
            root_folder = name.split('/')[0]
            root_folders.add(root_folder)
        assert len(root_folders) == 1,\
            'This archive does not have exactly one folder at the top level; are you sure it\'s a Torch model file?'
        root_folder = next(iter(root_folders))

        zipf.write(metadata_tmp_file_abs,
                   root_folder + '/' + destination_path,
                   compresslevel=9,
                   compress_type=zipfile.ZIP_DEFLATED)

    try:
        os.remove(metadata_tmp_file_abs)
    except Exception as e:
        print('Warning: error deleting file {}: {}'.format(metadata_tmp_file_abs,str(e)))

# ...def add_metadata_to_megadetector_model_file(...)


def read_metadata_from_megadetector_model_file(model_file,
                                               relative_path='megadetector_info.json',
                                               verbose=False):
    """
    Reads custom MegaDetector metadata from a modified MegaDetector model file.

    Args:
        model_file (str): The model filename to read, typically .pt (.zip is also sensible)
        relative_path (str, optional): The relative path within the main folder of the model
            archive from which we should read the metadata.  This is not relative to the root
            of the archive, it's relative to the one and only folder at the root of the archive
            (this is a PyTorch convention).

    Returns:
        object: whatever we read from the metadata file, always a dict in practice.  Returns
        None if we failed to read the specified metadata file.
    """

    with zipfile.ZipFile(model_file,'r') as zipf:

        # Torch doesn't like anything in the root folder of the zipfile, so we put
        # it in the one and only folder.
        names = zipf.namelist()
        root_folders = set()
        for name in names:
            root_folder = name.split('/')[0]
            root_folders.add(root_folder)
        if len(root_folders) != 1:
            print('Warning: this archive does not have exactly one folder at the top level; ' + \
                  'are you sure it\'s a Torch model file?')
            return None
        root_folder = next(iter(root_folders))

        metadata_file = root_folder + '/' + relative_path
        if metadata_file not in names:
            # This is the case for MDv5a and MDv5b
            if verbose:
                print('Warning: could not find metadata file {} in zip archive {}'.format(
                    metadata_file,os.path.basename(model_file)))
            return None

        try:
            path = zipfile.Path(zipf,metadata_file)
            contents = path.read_text()
            d = json.loads(contents)
        except Exception as e:
            print('Warning: error reading metadata from path {}: {}'.format(metadata_file,str(e)))
            return None

        return d

# ...def read_metadata_from_megadetector_model_file(...)


#%% Inference classes

default_compatibility_mode = 'classic'

# This is a useful hack when I want to verify that my test driver (md_tests.py) is
# correctly forcing a specific compatibility mode (I use "classic-test" in that case)
require_non_default_compatibility_mode = False

class PTDetector:
    """
    Class that runs a PyTorch-based MegaDetector model.
    """

    def __init__(self, model_path, detector_options=None, verbose=False):

        if verbose:
            print('Initializing PTDetector (verbose)')
        
        # Set up the import environment for this model, unloading previous
        # YOLO library versions if necessary.
        _initialize_yolo_imports_for_model(model_path,
                                           detector_options=detector_options,
                                           verbose=verbose)

        # Parse options specific to this detector family
        force_cpu = False
        use_model_native_classes = False
        compatibility_mode = default_compatibility_mode

        if detector_options is not None:

            if 'force_cpu' in detector_options:
                force_cpu = parse_bool_string(detector_options['force_cpu'])
            if 'use_model_native_classes' in detector_options:
                use_model_native_classes = parse_bool_string(detector_options['use_model_native_classes'])
            if 'compatibility_mode' in detector_options:
                if detector_options['compatibility_mode'] is None:
                    compatibility_mode = default_compatibility_mode
                else:
                    compatibility_mode = detector_options['compatibility_mode']

        if require_non_default_compatibility_mode:

            print('### DEBUG: requiring non-default compatibility mode ###')
            assert compatibility_mode != 'classic'
            assert compatibility_mode != 'default'

        preprocess_only = False
        if (detector_options is not None) and \
           ('preprocess_only' in detector_options) and \
           (detector_options['preprocess_only']):
            preprocess_only = True

        if verbose or (not preprocess_only):
            print('Loading PT detector with compatibility mode {}'.format(compatibility_mode))

        model_metadata = read_metadata_from_megadetector_model_file(model_path)

        #: Image size passed to the letterbox() function; 1280 means "1280 on the long side, preserving
        #: aspect ratio".
        if model_metadata is not None and 'image_size' in model_metadata:
            self.default_image_size = model_metadata['image_size']
            if verbose:
                print('Loaded image size {} from model metadata'.format(self.default_image_size))
        else:
            self.default_image_size = 1280

        #: Either a string ('cpu','cuda:0') or a torch.device()
        self.device = 'cpu'

        #: Have we already printed a warning about using a non-standard image size?
        #:
        #: :meta private:
        self.printed_image_size_warning = False

        #: If this is False, we assume the underlying model is producing class indices in the
        #: set (0,1,2) (and we assert() on this), and we add 1 to get to the backwards-compatible
        #: MD classes (1,2,3) before generating output.  If this is True, we use whatever
        #: indices the model provides
        self.use_model_native_classes = use_model_native_classes

        #: This allows us to maintain backwards compatibility across a set of changes to the
        #: way this class does inference.  Currently should start with either "default" or
        #: "classic".
        self.compatibility_mode = compatibility_mode

        #: Stride size passed to YOLOv5's letterbox() function
        self.letterbox_stride = 32

        if 'classic' in self.compatibility_mode:
            self.letterbox_stride = 64

        #: Use half-precision inference... fixed by the model, generally don't mess with this
        self.half_precision = False

        if preprocess_only:
            return

        if not force_cpu:
            if torch.cuda.is_available():
                self.device = torch.device('cuda:0')
            try:
                if torch.backends.mps.is_built and torch.backends.mps.is_available():
                    self.device = 'mps'
            except AttributeError:
                pass
        try:
            self.model = PTDetector._load_model(model_path,
                                                device=self.device,
                                                compatibility_mode=self.compatibility_mode)

        except Exception as e:
            # In a very esoteric scenario where an old version of YOLOv5 is used to run
            # newer models, we run into an issue because the "Model" class became
            # "DetectionModel".  New YOLOv5 code handles this case by just setting them
            # to be the same, so doing that externally doesn't seem *that* rude.
            if "Can't get attribute 'DetectionModel'" in str(e):
                print('Forward-compatibility issue detected, patching')
                from models import yolo
                yolo.DetectionModel = yolo.Model
                self.model = PTDetector._load_model(model_path,
                                                    device=self.device,
                                                    compatibility_mode=self.compatibility_mode,
                                                    verbose=verbose)
            else:
                raise
        if (self.device != 'cpu'):
            if verbose:
                print('Sending model to GPU')
            self.model.to(self.device)


    @staticmethod
    def _load_model(model_pt_path, device, compatibility_mode='', verbose=False):

        if verbose:
            print(f'Using PyTorch version {torch.__version__}')

        # There are two very slightly different ways to load the model, (1) using the
        # map_location=device parameter to torch.load and (2) calling .to(device) after
        # loading the model.  The former is what we did for a zillion years, but is not
        # supported on Apple silicon at of 2029.09.  Switching to the latter causes
        # very slight changes to the output, which always make me nervous, so I'm not
        # doing a wholesale swap just yet.  Instead, we'll just do this on M1 hardware.
        if 'classic' in compatibility_mode:
            use_map_location = (device != 'mps')
        else:
            use_map_location = False

        if use_map_location:
            try:
                checkpoint = torch.load(model_pt_path, map_location=device, weights_only=False)
            # For a transitional period, we want to support torch 1.1x, where the weights_only
            # parameter doesn't exist
            except Exception as e:
                if "'weights_only' is an invalid keyword" in str(e):
                    checkpoint = torch.load(model_pt_path, map_location=device)
                else:
                    raise
        else:
            try:
                checkpoint = torch.load(model_pt_path, weights_only=False)
            # For a transitional period, we want to support torch 1.1x, where the weights_only
            # parameter doesn't exist
            except Exception as e:
                if "'weights_only' is an invalid keyword" in str(e):
                    checkpoint = torch.load(model_pt_path)
                else:
                    raise

        # Compatibility fix that allows us to load older YOLOv5 models with
        # newer versions of YOLOv5/PT
        for m in checkpoint['model'].modules():
            t = type(m)
            if t is torch.nn.Upsample and not hasattr(m, 'recompute_scale_factor'):
                m.recompute_scale_factor = None

        if use_map_location:
            model = checkpoint['model'].float().fuse().eval()
        else:
            model = checkpoint['model'].float().fuse().eval().to(device)

        return model

    # ...def _load_model(...)


    def generate_detections_one_image(self,
                                      img_original,
                                      image_id='unknown',
                                      detection_threshold=0.00001,
                                      image_size=None,
                                      skip_image_resizing=False,
                                      augment=False,
                                      preprocess_only=False,
                                      verbose=False):
        """
        Applies the detector to an image.

        Args:
            img_original (Image): the PIL Image object (or numpy array) on which we should run the
                detector, with EXIF rotation already handled
            image_id (str, optional): a path to identify the image; will be in the "file" field
                of the output object
            detection_threshold (float, optional): only detections above this confidence threshold
                will be included in the return value
            image_size (tuple, optional): image size to use for inference, only mess with this if
                (a) you're using a model other than MegaDetector or (b) you know what you're getting into
            skip_image_resizing (bool, optional): whether to skip internal image resizing (and rely on
                external resizing), only mess with this if (a) you're using a model other than MegaDetector
                or (b) you know what you're getting into
            augment (bool, optional): enable (implementation-specific) image augmentation
            preprocess_only (bool, optional): only run preprocessing, and return the preprocessed image
            verbose (bool, optional): enable additional debug output

        Returns:
            dict: a dictionary with the following fields:
                - 'file' (filename, always present)
                - 'max_detection_conf' (removed from MegaDetector output files by default, but generated here)
                - 'detections' (a list of detection objects containing keys 'category', 'conf', and 'bbox')
                - 'failure' (a failure string, or None if everything went fine)
        """

        result = {'file': image_id }
        detections = []
        max_conf = 0.0

        if preprocess_only:
            assert 'classic' in self.compatibility_mode, \
                'Standalone preprocessing only supported in "classic" mode'
            assert not skip_image_resizing, \
                'skip_image_resizing and preprocess_only are exclusive'

        if detection_threshold is None:

            detection_threshold = 0

        try:

            # If the caller wants us to skip all the resizing operations...
            if skip_image_resizing:

                if isinstance(img_original,dict):
                    image_info = img_original
                    img = image_info['img_processed']
                    scaling_shape = image_info['scaling_shape']
                    letterbox_pad = image_info['letterbox_pad']
                    letterbox_ratio = image_info['letterbox_ratio']
                    img_original = image_info['img_original']
                    img_original_pil = image_info['img_original_pil']
                else:
                    img = img_original

            else:

                img_original_pil = None
                # If we were given a PIL image

                if not isinstance(img_original,np.ndarray):
                    img_original_pil = img_original
                    img_original = np.asarray(img_original)

                # PIL images are RGB already
                # img_original = img_original[:, :, ::-1]

                # Save the original shape for scaling boxes later
                scaling_shape = img_original.shape

                # If the caller is requesting a specific target size...
                if image_size is not None:

                    assert isinstance(image_size,int)

                    if not self.printed_image_size_warning:
                        print('Using user-supplied image size {}'.format(image_size))
                        self.printed_image_size_warning = True

                # Otherwise resize to self.default_image_size
                else:

                    image_size = self.default_image_size
                    self.printed_image_size_warning = False

                # ...if the caller has specified an image size

                # In "classic mode", we only do the letterboxing resize, we don't do an
                # additional initial resizing operation
                if 'classic' in self.compatibility_mode:

                    resize_ratio = 1.0

                # Resize the image so the long side matches the target image size.  This is not
                # letterboxing (i.e., padding) yet, just resizing.
                else:

                    use_ceil_for_resize = ('use_ceil_for_resize' in self.compatibility_mode)

                    h,w = img_original.shape[:2]
                    resize_ratio = image_size / max(h,w)

                    # Only resize if we have to
                    if resize_ratio != 1:

                        # Match what yolov5 does: use linear interpolation for upsizing;
                        # area interpolation for downsizing
                        if resize_ratio > 1:
                            interpolation_method = cv2.INTER_LINEAR
                        else:
                            interpolation_method = cv2.INTER_AREA

                        if use_ceil_for_resize:
                            target_w = math.ceil(w * resize_ratio)
                            target_h = math.ceil(h * resize_ratio)
                        else:
                            target_w = int(w * resize_ratio)
                            target_h = int(h * resize_ratio)

                        img_original = cv2.resize(
                            img_original, (target_w, target_h),
                            interpolation=interpolation_method)

                if 'classic' in self.compatibility_mode:

                    letterbox_auto = True
                    letterbox_scaleup = True
                    target_shape = image_size

                else:

                    letterbox_auto = False
                    letterbox_scaleup = False

                    # The padding to apply as a fraction of the stride size
                    pad = 0.5

                    model_stride = int(self.model.stride.max())

                    max_dimension = max(img_original.shape)
                    normalized_shape = [img_original.shape[0] / max_dimension,
                                        img_original.shape[1] / max_dimension]
                    target_shape = np.ceil(np.array(normalized_shape) * image_size / model_stride + \
                                           pad).astype(int) * model_stride

                # Now we letterbox, which is just padding, since we've already resized.
                img,letterbox_ratio,letterbox_pad = letterbox(img_original,
                                                              new_shape=target_shape,
                                                              stride=self.letterbox_stride,
                                                              auto=letterbox_auto,
                                                              scaleFill=False,
                                                              scaleup=letterbox_scaleup)

                if preprocess_only:

                    assert 'file' in result
                    result['img_processed'] = img
                    result['img_original'] = img_original
                    result['img_original_pil'] = img_original_pil
                    result['target_shape'] = target_shape
                    result['scaling_shape'] = scaling_shape
                    result['letterbox_ratio'] = letterbox_ratio
                    result['letterbox_pad'] = letterbox_pad
                    return result

            # ...are we doing resizing here, or were images already resized?

            # Convert HWC to CHW (which is what the model expects).  The PIL Image is RGB already,
            # so we don't need to mess with the color channels.
            #
            # TODO, this could be moved into the preprocessing loop

            img = img.transpose((2, 0, 1)) # [::-1]
            img = np.ascontiguousarray(img)
            img = torch.from_numpy(img)
            img = img.to(self.device)
            img = img.half() if self.half_precision else img.float()
            img /= 255

            # In practice this is always true
            if len(img.shape) == 3:
                img = torch.unsqueeze(img, 0)

            # Run the model
            pred = self.model(img,augment=augment)[0]

            if 'classic' in self.compatibility_mode:
                nms_conf_thres = detection_threshold
                nms_iou_thres = 0.45
                nms_agnostic = False
                nms_multi_label = False
            else:
                nms_conf_thres = detection_threshold # 0.01
                nms_iou_thres = 0.6
                nms_agnostic = False
                nms_multi_label = True

            # As of PyTorch 1.13.0.dev20220824, nms is not implemented for MPS.
            #
            # Send predictions back to the CPU for NMS.
            if self.device == 'mps':
                pred_nms = pred.cpu()
            else:
                pred_nms = pred

            # NMS
            pred = non_max_suppression(prediction=pred_nms,
                                       conf_thres=nms_conf_thres,
                                       iou_thres=nms_iou_thres,
                                       agnostic=nms_agnostic,
                                       multi_label=nms_multi_label)

            # In practice this is [w,h,w,h] of the original image
            gn = torch.tensor(scaling_shape)[[1, 0, 1, 0]]

            if 'classic' in self.compatibility_mode:

                ratio = None
                ratio_pad = None

            else:

                # letterbox_pad is a 2-tuple specifying the padding that was added on each axis.
                #
                # ratio is a 2-tuple specifying the scaling that was applied to each dimension.
                #
                # The scale_boxes function expects a 2-tuple with these things combined.
                ratio = (img_original.shape[0]/scaling_shape[0], img_original.shape[1]/scaling_shape[1])
                ratio_pad = (ratio, letterbox_pad)

            # This is a loop over detection batches, which will always be length 1 in our case,
            # since we're not doing batch inference.
            #
            # det = pred[0]
            #
            # det is a torch.Tensor with size [nBoxes,6].  In practice the boxes are sorted
            # in descending order by confidence.
            #
            # Columns are:
            #
            # x0,y0,x1,y1,confidence,class
            #
            # At this point, these are *non*-normalized values, referring to the size at which we
            # ran inference (img.shape).
            for det in pred:

                if len(det) == 0:
                    continue

                # Rescale boxes from img_size to im0 size, and undo the effect of padded letterboxing
                if 'classic' in self.compatibility_mode:

                    det[:, :4] = scale_coords(img.shape[2:], det[:, :4], img_original.shape).round()

                else:
                    # After this scaling, each element of det is a box in x0,y0,x1,y1 format, referring to the
                    # original pixel dimension of the image, followed by the class and confidence
                    det[:, :4] = scale_coords(img.shape[2:], det[:, :4], scaling_shape, ratio_pad).round()

                # Loop over detections
                for *xyxy, conf, cls in reversed(det):

                    if conf < detection_threshold:
                        continue

                    # Convert this box to normalized cx, cy, w, h (i.e., YOLO format)
                    xywh = (xyxy2xywh(torch.tensor(xyxy).view(1, 4)) / gn).view(-1).tolist()

                    # Convert from normalized cx/cy/w/h (i.e., YOLO format) to normalized
                    # left/top/w/h (i.e., MD format)
                    api_box = ct_utils.convert_yolo_to_xywh(xywh)

                    if 'classic' in self.compatibility_mode:
                        api_box = ct_utils.truncate_float_array(api_box, precision=COORD_DIGITS)
                        conf = ct_utils.truncate_float(conf.tolist(), precision=CONF_DIGITS)
                    else:
                        api_box = ct_utils.round_float_array(api_box, precision=COORD_DIGITS)
                        conf = ct_utils.round_float(conf.tolist(), precision=CONF_DIGITS)

                    if not self.use_model_native_classes:
                        # The MegaDetector output format's categories start at 1, but all YOLO-based
                        # MD models have category numbers starting at 0.
                        cls = int(cls.tolist()) + 1
                        if cls not in (1, 2, 3):
                            raise KeyError(f'{cls} is not a valid class.')
                    else:
                        cls = int(cls.tolist())

                    detections.append({
                        'category': str(cls),
                        'conf': conf,
                        'bbox': api_box
                    })
                    max_conf = max(max_conf, conf)

                # ...for each detection in this batch

            # ...for each detection batch (always one iteration)

        # ...try

        except Exception as e:

            result['failure'] = FAILURE_INFER
            print('PTDetector: image {} failed during inference: {}\n'.format(image_id, str(e)))
            # traceback.print_exc(e)
            print(traceback.format_exc())

        result['max_detection_conf'] = max_conf
        result['detections'] = detections

        return result

    # ...def generate_detections_one_image(...)

# ...class PTDetector


#%% Command-line driver

# For testing only... you don't really want to run this module directly.

if __name__ == '__main__':

    pass

    #%%

    import os #noqa
    from megadetector.visualization import visualization_utils as vis_utils

    model_file = os.environ['MDV5A']
    im_file = os.path.expanduser('~/git/MegaDetector/images/nacti.jpg')

    detector = PTDetector(model_file)
    image = vis_utils.load_image(im_file)

    res = detector.generate_detections_one_image(image, im_file, detection_threshold=0.00001)
    print(res)
